# -*- coding: utf-8 -*-
#
# @project TirjaPy
# @file src/tirjapy/utils/Porygon.py
# @author  Shreos Roychowdhury <shreos@tirja.com>
# @version 1.0.0
# 
# @section DESCRIPTION
# 
#   Porygon.py :  handle porygon string
# 
# @section LICENSE
# 
# Copyright (c) 2025 Shreos Roychowdhury.
# Copyright (c) 2025 Tirja Consulting LLP.
# 
# This source code is protected under international copyright law.  All rights
# reserved and protected by the copyright holders.
#
# This file is confidential and only available to authorized individuals with the
# permission of the copyright holders.  If you encounter this file and do not have
# permission, please contact the copyright holders and delete this file.
#
# THE SOFTWARE IS PROVIDED , WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# 

import io
import itertools
import math
from typing import List, Tuple

def _pcitr(iterable):
	return zip(iterable, itertools.islice(iterable, 1, None))

def _py2_round(x):
	# The polyline algorithm uses Python 2's way of rounding
	return int(math.copysign(math.floor(math.fabs(x) + 0.5), x))

def _write(output, curr_value, prev_value, factor):
	curr_value = _py2_round(curr_value * factor)
	prev_value = _py2_round(prev_value * factor)
	coord = curr_value - prev_value
	coord <<= 1
	coord = coord if coord >= 0 else ~coord

	while coord >= 0x20:
		output.write(chr((0x20 | (coord & 0x1f)) + 63))
		coord >>= 5

	output.write(chr(coord + 63))

def _trans(value, index):
	byte, result, shift = None, 0, 0

	comp = None
	while byte is None or byte >= 0x20:
		byte = ord(value[index]) - 63
		index += 1
		result |= (byte & 0x1f) << shift
		shift += 5
		comp = result & 1

	return ~(result >> 1) if comp else (result >> 1), index

class Porygon:
	""" class def """

	def decode(self, items: int, expression: str, precision: int = 5) -> List[Tuple]:
		"""
		Decode a polyline string into a set of cvalues.

		:param items : No of items 
		:param expression: Polyline string, e.g. '_ih_pfA_hmFoezGwwqa@{msh@cneqP_qo]gyuC??s_L?_{rc@~uJ???_jP'
		:param precision: Precision of the encoded cvalues.  The default value is 5.
		:return: List of coordinate tuples in (ts, a, b, c, d, e) order
		"""
		cvalues, index, length, factor = [], 0, len(expression), float(10 ** precision)
		xitem = [0.0] * items

		while index < length:
			yitem = [0.0] * items
			for x in range(items):
				xx_change, index = _trans(expression, index)
				xitem[x] += xx_change
				yitem[x] = xitem[x]/factor

			cvalues.append(tuple(yitem))

		return cvalues

	def encode(self, items: int, cvalues: List[Tuple], precision: int = 5) -> str:
		"""
		Encode a set of cvalues in a polyline string.

		:param items : No of items 
		:param cvalues: List of cvalue tuples
		:param precision: Precision of the cvalues to encode.  The default value is 5.
		:return: The encoded polyline string.
		"""
		output, factor = io.StringIO(), int(10 ** precision)

		for x in range(items):
			_write(output, cvalues[0][x], 0, factor)

		for prev, curr in _pcitr(cvalues):
			for x in range(items):
				_write(output, curr[x], prev[x], factor)

		return output.getvalue()



# -*- coding: utf-8 -*-
#
# @project TirjaPy
# @file src/tirjapy/base/WebServiceBase.py
# @author  Shreos Roychowdhury <shreos@tirja.com>
# @version 1.0.0
# 
# @section DESCRIPTION
# 
#   WebServiceBase.py : base for Web Service
# 
# @section LICENSE
# 
# Copyright (c) 2025 Shreos Roychowdhury.
# Copyright (c) 2025 Tirja Consulting LLP.
# 
# This source code is protected under international copyright law.  All rights
# reserved and protected by the copyright holders.
#
# This file is confidential and only available to authorized individuals with the
# permission of the copyright holders.  If you encounter this file and do not have
# permission, please contact the copyright holders and delete this file.
#
# THE SOFTWARE IS PROVIDED , WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# 

import os
import sys
import re
import json
import random
import uuid
import base64
import string
from pathlib import Path

from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad

class WebServiceBase():
	"""Default Simple WebService base class."""

	def _GetPath(self, htpath):
		return htpath if htpath.startswith('/') else Path(Path.cwd(),htpath).as_posix()

	def _RequiredField(self, data, field, blank_ok=False):
		""" init handler required and field from data"""
		if field in data and len(str(data[field])) > 0:
			return data[field]
		elif field in data and blank_ok:
			return data[field]
		raise ValueError("Missing Field: " + field)

	def _RequiredInteger(self, data, field, blank_ok=True):
		""" init handler required int field from data"""
		if field in data and len(str(data[field])) > 0:
			return int(data[field])
		elif field in data and blank_ok:
			return 0
		raise ValueError("Missing Integer: " + field)

	def _RequiredFloat(self, data, field, blank_ok=True):
		""" init handler required float field from data"""
		if field in data and len(str(data[field])) > 0:
			return float(data[field])
		elif field in data and blank_ok:
			return 0.0
		raise ValueError("Missing Float: " + field)

	def _RequiredArray(self, data, field):
		""" init handler required for array field from data"""
		if field in data and type(data[field]) is list:
			return data[field]
		raise ValueError("Missing Array Field: " + field)

	def _OptionalField(self, data, field , default=''):
		""" init handler optional and field from data"""
		if field in data and len(str(data[field])) > 0:
			return data[field]
		return default

	def _OptionalInteger(self, data, field , default=0):
		""" init handler optional int field from data"""
		if field in data and len(str(data[field])) > 0:
			return int(data[field]) 
		return default

	def _OptionalFloat(self, data, field , default=0.0):
		""" init handler optional float field from data"""
		if field in data and len(str(data[field])) > 0:
			return float(data[field])
		return default

	def _OptionalArray(self, data, field, default=[]):
		""" init handler for array field from data"""
		if field in data and type(data[field]) is list:
			return data[field]
		return default

	def _HandlePost(self, request):
		""" get post content"""
		content_type = request.headers.get('Content-Type')
		if content_type == 'application/json':
			return request.json
		elif re.match('multipart/form-data',content_type) or re.match('application/x-www-form-urlencoded',content_type):
			robj = {}
			for item in request.form.keys():
				robj[item]=request.form[item]
			return robj
		else:
			return {}

	def _EncodeDict(self, skey, data):
		usekey=base64.b64decode(skey)
		BS=AES.block_size
		iv = ''.join(random.choices(string.ascii_lowercase + string.digits, k = BS))
		# aes = AES.new(key, AES.MODE_CBC, bytes(iv,'utf8'))
		aes = AES.new(usekey, AES.MODE_CBC, iv.encode('utf8'))
		xdata = json.dumps(data)
		pad = BS - len(xdata) % BS
		xdata = xdata + pad * chr(pad)
		encrypted = aes.encrypt(xdata.encode('utf8'))
		return iv+base64.urlsafe_b64encode(encrypted).decode('utf8')

	def _DecodeDict(self, skey, data):
		output = ""
		try:
			usekey=base64.b64decode(skey)
			BS=AES.block_size
			iv = data[:BS]
			cipher = AES.new(usekey, AES.MODE_CBC, iv.encode('utf8') )
			output = unpad(cipher.decrypt( base64.urlsafe_b64decode(data[BS:]) ), BS).decode('utf-8')
		except:
			output = "{}"

		return json.loads(output)

	def _PopulateBasicHeaderCreds(self, headers, username, passwd):
		""" function to populate username password in headers """

		if not username or not passwd:
			return headers
		use_creds = ':'.join([username,passwd])
		base64creds = base64.b64encode(use_creds.encode('ascii'))
		headers['Authorization'] = "Basic {}".format(base64creds.decode('ascii'))

	def _SplitCredsFx(self, creds):
		""" function to server, username, passwd from creds for backward compat """
		server = self._RequiredField( creds, 'server' )
		username = self._OptionalField( creds, 'username' )
		passwd = self._OptionalField( creds, 'passwd' )
		return ( server , username, passwd )

	def _PostJsonData(self, creds, uri, data):
		""" post file as json data """

		## pre-flight
		server, username, passwd = self._SplitCredsFx(creds)
		url = '/'.join([server, uri])

		## handle headers
		headers = {}
		self._PopulateBasicHeaderCreds(headers, username, passwd)
		headers['Content-Type'] = 'application/json'

		output = { 'error' : True , 'status' : 'unknown' }
		r = self.http.request( 'POST', url, headers=headers, body=json.dumps(data))
		# print(r.status)

		if r.status==200:
			output = json.loads( r.data.decode('utf-8') )
		return output

	def _PostFormData(self, creds, uri, fields):
		""" post file as json data """

		## pre-flight
		server, username, passwd = self._SplitCredsFx(creds)
		url = '/'.join([server, uri])

		## handle headers
		headers = {}
		self._PopulateBasicHeaderCreds(headers, username, passwd)

		output = { 'error' : True , 'status' : 'unknown' }
		r = self.http.request( 'POST', url, headers=headers, fields=fields, encode_multipart=True)

		if r.status==200:
			output = json.loads( r.data.decode('utf-8') )
		return output

	def _GetJsonData(self, creds, uri, fields={}):
		""" get file as json data """

		## pre-flight
		server, username, passwd = self._SplitCredsFx(creds)
		url = '/'.join([server, uri])

		## handle headers
		headers = {}
		self._PopulateBasicHeaderCreds(headers, username, passwd)
		output = { 'error' : True , 'status' : 'unknown' }
		r = self.http.request( 'GET', url, headers=headers, fields=fields)

		if r.status==200:
			output = json.loads( r.data.decode('utf-8') )
		return output




# -*- coding: utf-8 -*-
#
# @project TirjaPy
# @file src/tirjapy/utils/FileStoreBase.py
# @author  Shreos Roychowdhury <shreos@tirja.com>
# @version 1.0.0
# 
# @section DESCRIPTION
# 
#   FileStoreBase.py :  File Store Base
# 
# @section LICENSE
# 
# Copyright (c) 2025 Shreos Roychowdhury.
# Copyright (c) 2025 Tirja Consulting LLP.
# 
# This source code is protected under international copyright law.  All rights
# reserved and protected by the copyright holders.
#
# This file is confidential and only available to authorized individuals with the
# permission of the copyright holders.  If you encounter this file and do not have
# permission, please contact the copyright holders and delete this file.
#
# THE SOFTWARE IS PROVIDED , WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# 

import os
import json
import math
import time
from enum import Enum, IntEnum
from tirjapy.utils.TypedEnum import TypedEnum

from tirjapy.utils.HandleQuotes import HandleQuotes
from tirjapy.base.StoreBase import StoreBase
from tirjapy.utils.StorageHandle import StorageHandle
from tirjapy.utils.CalyrexHandle import CalyrexHandle

class SyncTypes(metaclass=TypedEnum):
	LOCAL                       = 'local'             # Local only
	FETCHED                     = 'fetched'           # Fetched from S3
	WRITTEN                     = 'written'           # Written to S3

class ModelPath(HandleQuotes):
	"""Default model path class """

	def __init__(self , xdata):
		""" init fx inits handles """
		self.upload_folder  = self._RequiredField(xdata, 'upload_folder')
		self.s3_bucket      = self._RequiredField(xdata, 's3_bucket')
		self.work_path      = self._MicroS3Path()

	def GetData(self, blank=False):
		""" Prints Json of class """
		return {
			's3_bucket'  : self.s3_bucket,
			'work_path' : self.work_path,
			'upload_folder' : self.upload_folder,
		}

class FileStoreBase(StoreBase):

	def __init__(self):
		""" default constructor """

		# - File File Tags
		self.s3_bucket            = ''                  # String.       S3 Bucket 
		self.s3_path              = ''                  # String.       S3 Path for file
		self.local_path           = ''                  # String.       Local Path for file

		self.clean                = False               # Bool.         If set file is cleaned, default False
		self.fetched              = False               # Bool.         If set file is fetched
		self.sync_type            = SyncTypes.LOCAL     # Enum.         Sync State for file

	# Functions

	def GetData(self, blank=False):
		""" getter for json """
		if blank and len(self.s3_bucket)==0 and len(self.s3_path)==0 and len(self.local_path)==0:
			return {}
		return {
			's3_bucket' : self.s3_bucket,
			's3_path' : self.s3_path,
			'local_path' : self.local_path,
			'sync_type' : self.sync_type
		}

	def InitData(self, data, file_name):
		""" init for using MakePptxStore and file_name"""
		self.s3_bucket = data.s3_bucket
		self.s3_path = '/'.join([data.work_path, file_name])
		self.local_path = '/'.join([data.upload_folder, self.s3_bucket, self.s3_path])
		dirname = os.path.dirname(self.local_path)
		os.makedirs(dirname,exist_ok = True)

	def WriteLocal(self, file_obj, overwrite=False):
		""" Writes to local, overwrite"""
		if not overwrite and os.path.exists(self.local_path):
			raise ValueError(f"File exists at {self.local_path}")
		dirname = os.path.dirname(self.local_path)
		os.makedirs(dirname,exist_ok = True)
		file_obj.seek(0)
		file_obj.save(self.local_path)

	def WriteToS3(self, overwrite=False):
		""" Writes to S3"""
		if self.fetched and (not overwrite):
			return None
		shandle = StorageHandle()
		shandle.WriteFileObj(self)
		self.sync_type=SyncTypes.WRITTEN
		self.fetched=True
		
	def _FetchFromS3(self , overwrite=False):
		""" Reads from S3"""
		if (not overwrite) and os.path.isfile(self.local_path):
			return None
		dirname = os.path.dirname(self.local_path)
		os.makedirs(dirname,exist_ok = True)
		shandle = StorageHandle()
		shandle.ReadFileObj(self)
		self.sync_type=SyncTypes.FETCHED
		self.fetched=True

	def GetLocalPath(self):
		""" Get Local file path """
		if len(self.local_path)==0:
			raise ValueError("local_path is blank")
		return self.local_path

	def LocalRequired(self , overwrite=False):
		""" Get Local file, if not fetch from S3 Reads from S3"""
		self._FetchFromS3(overwrite)
		return self.local_path

	def LocalOptional(self , overwrite=False):
		""" Ignore blank, Get Local file, if not fetch from S3 Reads from S3"""
		if len(self.local_path)>0:
			return self.LocalRequired(overwrite)
		return ''

	def GetCalyrexURL(self, user=''):
		""" Gets Calyrex URL"""
		chandle = CalyrexHandle()
		return chandle.GetURL( user, self.s3_bucket, self.s3_path)
		
	def CleanUp(self):
		""" Deletes local file"""
		if self.local_path and self.clean:
			self._CleanUpFile( self._GetPath(self.local_path))

